SystemOrganization addCategory: #'Rome-Reference'!

RomeCanvas subclass: #RomeReferenceCanvas
	instanceVariableNames: 'stack ctm pen deviceFill devicePen deviceClip deviceTolerance path capStyle joinStyle miterLimit dashSpec dashOffset font fontMatrix noPen noFill noClip'
	classVariableNames: 'DefaultPen'
	poolDictionaries: ''
	category: 'Rome-Reference'!

!RomeReferenceCanvas commentStamp: '<historical>' prior: 0!
A canvas is a two-dimensional medium on which objects are drawn in a device-independent manner. Canvases keep track of various attributes related to the drawing operations.

Instance variables:
	stack		<RomeCanvas>	- The internal stack.
	ctm 		<CDisplayTransform>	- The current user to device space transformation.
	
	deviceFill		<RomeFill>			- The current fill.
	devicePen		<RomePen>			- The current pen.
	deviceClip		<RomeClipshape>		- The current clip shape
	deviceTolerance	<Number>			- The tolerance when converting (stored as square!!)

	capStyle 	<Symbol>	- The cap style for lines, #butt, #square, or #round.
	joinStyle 	<Symbol>	- The join style for lines, #miter, #bevel, or #round.
	miterLimit 	<Number>	- The miter limit for #miter join style.
	dashSpec 	<Array of: Number> - The line dash specification.

	font 		<Font>		- The currently selected font.
	fontName	<String>		- The font name.
	fontFlags	<Number>	- Font emphasis flags.
	fontSize		<Number>	- Font size (in user space units).
	fontMatrix	<CDisplayTransform>	- The transformation from design to user space units.
	fontColor	<Color>		- The current font color.

	noPen 		<CPen>		- The "null" pen
	noFill		<CFill>		- The "null" fill

Reference:
A canvas keeps a set of attributes and draws shapes. The shapes are filled with the current fill, outlined by the current pen and clipped against the current clip shape. All shapes can be converted into paths and from there (implicitly and lossy) into polygons/polylines. A shape is drawn using the following stages:

    * Shape generalization: As the first step we determine whether the current transformation is shape-preserving or not. If it is not shape-preserving we generalize the input to an appropriate path.
    * Transformation: The output from the first step is transformed by the current transformation (e.g., converted into device coordinates). In addition, fill, and clip shape are transformed into device coordinates. At the end of this step there are no more transformations necessary.
    * Clipping: The output from the previous step is clipped against the current clip shape.
    * Filling: The output from the previous step is filled with the current fill. If the fill cannot be handled by the canvas directly it will ask the fill to draw the shape accordingly. The fill can then use the appropriate fallback strategies for drawing itself.

When a pen is specified to stroke the shape accordingly, the process is conceptually equivalent to the above (though we will typically see vast optimizations here) merely by filling the shape computed as the Minkowski sum of both pen outline and path outline. 
!

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:49'!
exampleArc
	"self exampleArc"
	| canvas center radius angle1 angle2 |
	canvas := self newExampleCanvas.
	"Draw arc"
	center := 0.5@0.5.
	radius := 0.4.
	angle1 := 45.
	angle2 := 180.

	canvas drawArc: radius at: center from: angle1 to: angle2.
	"Draw helper dot"
	canvas selectNoPen.
	canvas selectFill: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawArc: 0.05 at: center from: 0 to: 360.
	"Draw helper lines"
	canvas selectPenWidth: 0.03 color: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawPath:[
		canvas arc: radius at: center from: angle1 to: angle1. "no-op, just positioning"
		canvas lineTo: center.
	].
	canvas drawPath:[
		canvas arc: radius at: center from: angle2 to: angle2. "no-op, just positioning"
		canvas lineTo: center.
	].
	canvas finish.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:49'!
exampleArcReverse
	"self exampleArcReverse"
	| canvas center radius angle1 angle2 |
	canvas := self newExampleCanvas.
	"Draw arc"
	center := 0.5@0.5.
	radius := -0.4.
	angle1 := 45.
	angle2 := 180.

	canvas drawArc: radius at: center from: angle1 to: angle2.
	"Draw helper dot"
	canvas selectNoPen.
	canvas selectFill: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawArc: 0.05 at: center from: 0 to: 360.
	"Draw helper lines"
	canvas selectPenWidth: 0.03 color: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawPath:[
		canvas arc: radius at: center from: angle1 to: angle1. "no-op, just positioning"
		canvas lineTo: center.
	].
	canvas drawPath:[
		canvas arc: radius at: center from: angle2 to: angle2. "no-op, just positioning"
		canvas lineTo: center.
	].

! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleClip
	"self exampleClip"
	| canvas |
	canvas := self newExampleCanvas.
	canvas clipPath:[canvas arc: 0.3 at: 0.5@0.5 from: 0 to: 360; closePath].
	canvas selectFill: Color blue.
	canvas drawRectangle: (0@0 corner: 1@1).
	canvas selectPenWidth: 0.03 color: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawLineFrom: 0@0 to: 1@1.
	canvas drawLineFrom: 0@1 to: 1@0.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 22:54'!
exampleClipImage
	"self exampleClipImage"
	| canvas image |
	canvas := self newExampleCanvas.
	canvas clipPath:[
		canvas arc: 0.3 at: 0.5@0.5 from: 0 to: 360.
		canvas closePath.
	].
	image := self newExampleImage.
	canvas scaleBy: 1.0 / image extent.
	canvas drawImage: image at: 0@0.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleClipRect
	"RomeBalloonCanvas exampleClipRect"
	| canvas rect |
	canvas := self newExampleCanvas.
	rect := 0.25@0.25 corner: 0.75@0.75.
	canvas clipRectangle: rect.
	canvas drawRectangle: rect.
	canvas selectPenWidth: 0.03 color: (Color r: 1.0 g: 0.2 b: 0.2).
	canvas drawLineFrom: 0@0 to: 1@1.
	canvas finish.! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleCurveTo
	"self exampleCurveTo"
	| canvas p0 p1 p2 p3 |
	p0 := 0.1@0.5.
	p1 := 0.4@0.9.
	p2 := 0.6@0.1.
	p3 := 0.9@0.5.
	canvas := self newExampleCanvas.
	canvas drawCurveFrom: p0 via: p1 and: p2 to: p3.
	canvas selectPenWidth: 0.03 color: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawLineFrom: p0 to: p1.
	canvas drawLineFrom: p2 to: p3.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleCurveTo2
	"self exampleCurveTo2"
	| canvas p0 p1 p2 p3 |
	canvas := self newExampleCanvas.
	canvas dashSpec: #(
		0.01		"ink"
		0.01		"skip"
	).
	canvas selectPenWidth: 0.005 color: Color black.
	p0 := 0.1@0.5.
	p1 := 0.4@0.9.
	p2 := 0.6@0.1.
	p3 := 0.9@0.5.
	canvas drawCurveFrom: p0 via: p1 and: p2 to: p3.
	canvas selectPenWidth: 0.005 color: (Color r: 1.0 g: 0.2 b: 0.2 alpha: 0.6).
	canvas drawLineFrom: p0 to: p1.
	canvas drawLineFrom: p2 to: p3.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 22:34'!
exampleDashes
	"self exampleDashes"
	| canvas |
	canvas := self newExampleCanvas.
	canvas dashSpec: #(
		0.20		"ink"
		0.05		"skip"
		0.05		"ink"
		0.05		"skip"
	).
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 22:36'!
exampleDashes2
	"self exampleDashes2"
	| canvas |
	canvas := self newExampleCanvas.
	canvas dashSpec: #(
		0.20		"ink"
		0.05		"skip"
		0.05		"ink"
		0.05		"skip"
	).
	canvas drawPath:[
		canvas moveTo: 0.3@0.3.
		canvas lineBy: 0.35@0.0.
		canvas lineBy: 0.0@0.35.
		canvas lineBy: -0.35@0.0.
		canvas lineBy: 0.0@-0.35.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleDashes3
	"self exampleDashes3"
	| canvas |
	canvas := self newExampleCanvas.
	canvas dashSpec: #(
		0.01		"ink"
		0.01		"skip"
	).
	canvas selectPenWidth: 0.005 color: Color black.
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
		canvas closePath.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 20:38'!
exampleFillAndStroke
	"self exampleFillAndStroke"
	| canvas |
	canvas := self newExampleCanvas.
	canvas selectFill: Color blue.
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
		canvas closePath.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 20:38'!
exampleFillAndStroke2
	"self exampleFillAndStroke2"
	| canvas |
	canvas := self newExampleCanvas.
	canvas selectFill: Color blue.
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
		canvas closePath.

		canvas moveTo: 0.25@0.1.
		canvas lineBy: 0.2@0.2.
		canvas lineBy: -0.2@0.2.
		canvas lineBy: -0.2@-0.2.
		canvas closePath.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 23:00'!
exampleFillAndStroke3
	"self exampleFillAndStroke3"
	| canvas |
	canvas := self newExampleCanvas.
	canvas dashSpec: #(
		0.1		"ink"
		0.1		"skip"
	).
	canvas selectFill: Color blue.
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
		canvas closePath.
	].
	canvas drawPath:[
		canvas moveTo: 0.25@0.1.
		canvas lineBy: 0.2@0.2.
		canvas lineBy: -0.2@0.2.
		canvas lineBy: -0.2@-0.2.
		canvas closePath.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleFillAndStroke4
	"self exampleFillAndStroke4"
	| canvas |
	canvas := self newExampleCanvas.
	canvas dashSpec: #(
		0.01		"ink"
		0.01		"skip"
	).
	canvas selectFill: (Color r: 0.5 g: 0.5 b: 1.0).
	canvas selectPenWidth: 0.005 color: Color black.
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
		canvas closePath.
	].
	canvas drawPath:[
		canvas moveTo: 0.25@0.1.
		canvas lineBy: 0.2@0.2.
		canvas lineBy: -0.2@0.2.
		canvas lineBy: -0.2@-0.2.
		canvas closePath.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 18:43'!
exampleGradient
	"self exampleGradient"
	| fill canvas |
	canvas := self newExampleCanvas.

	fill := canvas newLinearGradient.
	fill origin: 0@0.
	fill direction: 0@1.
	fill addColorStop: Color black at: 1.0.
	fill addColorStop: Color white at: 0.0.
	canvas fill: fill.
	canvas drawRectangle: (0@0 corner: 1@1).

	fill := canvas newRadialGradient.
	fill origin: 0.45@0.4.
	fill direction: -0.05@0.
	fill addColorStop: Color white at: 0.0.
	fill addColorStop: Color white at: 0.3.
	fill addColorStop: Color black at: 1.0.
	canvas fill: fill.
	canvas drawOval: (0.2@0.2 corner: 0.8@0.8).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:43'!
exampleImage
	"self exampleImage"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleImage.
	canvas scaleBy: 1.0 / image extent.
	canvas drawImage: image at: 0@0.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:46'!
exampleImage2
	"self exampleImage2"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleImage.
	canvas translateBy: 0.5@0.5.
	canvas rotateBy: 45.0.
	canvas scaleBy: 1.0 / image extent.
	canvas drawImage: image at: 0@0.! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:43'!
exampleImage3
	"self exampleImage3"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleImage.
	canvas drawImage: image in: (0.0@0.0 corner: 1.0@1.0).
	canvas drawImage: image in: (0.1@0.1 corner: 0.9@0.9).
	canvas drawImage: image in: (0.2@0.2 corner: 0.8@0.8).
	canvas drawImage: image in: (0.3@0.3 corner: 0.7@0.5).
	canvas drawImage: image in: (0.4@0.4 corner: 0.6@0.6).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:44'!
exampleImage4
	"self exampleImage4"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleImage.
	canvas drawImage: image in: (0.0@0.0 corner: 1.0@1.0).
	canvas rotateBy: 45.
	canvas drawImage: image in: (0.1@0.1 corner: 0.9@0.9).
	canvas rotateBy: 45.
	canvas drawImage: image in: (0.2@0.2 corner: 0.8@0.8).
	canvas rotateBy: 45.
	canvas drawImage: image in: (0.3@0.3 corner: 0.7@0.5).
	canvas rotateBy: 45.
	canvas drawImage: image in: (0.4@0.4 corner: 0.6@0.6).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:56'!
exampleImageAlpha
	"self exampleImageAlpha"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleAlphaImage.
	canvas scaleBy: 1.0 / image extent.
	canvas blendImage: image at: 0@0.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:57'!
exampleImageAlpha2
	"self exampleImageAlpha2"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleAlphaImage.
	canvas translateBy: 0.5@0.5.
	canvas rotateBy: 45.0.
	canvas scaleBy: 1.0 / image extent.
	canvas blendImage: image at: 0@0.! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:57'!
exampleImageAlpha3
	"self exampleImageAlpha3"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleAlphaImage.
	canvas blendImage: image in: (0.0@0.0 corner: 1.0@1.0).
	canvas blendImage: image in: (0.1@0.1 corner: 0.9@0.9).
	canvas blendImage: image in: (0.2@0.2 corner: 0.8@0.8).
	canvas blendImage: image in: (0.3@0.3 corner: 0.7@0.5).
	canvas blendImage: image in: (0.4@0.4 corner: 0.6@0.6).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:57'!
exampleImageAlpha4
	"self exampleImageAlpha4"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExampleAlphaImage.
	canvas blendImage: image in: (0.0@0.0 corner: 1.0@1.0).
	canvas rotateBy: 45.
	canvas blendImage: image in: (0.1@0.1 corner: 0.9@0.9).
	canvas rotateBy: 45.
	canvas blendImage: image in: (0.2@0.2 corner: 0.8@0.8).
	canvas rotateBy: 45.
	canvas blendImage: image in: (0.3@0.3 corner: 0.7@0.5).
	canvas rotateBy: 45.
	canvas blendImage: image in: (0.4@0.4 corner: 0.6@0.6).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:58'!
exampleImagePaint
	"self exampleImagePaint"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExamplePaintImage.
	canvas scaleBy: 1.0 / image extent.
	canvas paintImage: image at: 0@0.
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 20:04'!
exampleImagePaint2
	"self exampleImagePaint2"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExamplePaintImage.
	canvas translateBy: 0.5@0.5.
	canvas rotateBy: 45.0.
	canvas scaleBy: 1.0 / image extent.
	canvas paintImage: image at: 0@0.! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 20:04'!
exampleImagePaint3
	"self exampleImagePaint3"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExamplePaintImage.
	canvas paintImage: image in: (0.0@0.0 corner: 1.0@1.0).
	canvas paintImage: image in: (0.1@0.1 corner: 0.9@0.9).
	canvas paintImage: image in: (0.2@0.2 corner: 0.8@0.8).
	canvas paintImage: image in: (0.3@0.3 corner: 0.7@0.5).
	canvas paintImage: image in: (0.4@0.4 corner: 0.6@0.6).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 20:04'!
exampleImagePaint4
	"self exampleImage4"
	| canvas image |
	canvas := self newExampleCanvas.
	image := self newExamplePaintImage.
	canvas paintImage: image in: (0.0@0.0 corner: 1.0@1.0).
	canvas rotateBy: 45.
	canvas paintImage: image in: (0.1@0.1 corner: 0.9@0.9).
	canvas rotateBy: 45.
	canvas paintImage: image in: (0.2@0.2 corner: 0.8@0.8).
	canvas rotateBy: 45.
	canvas paintImage: image in: (0.3@0.3 corner: 0.7@0.5).
	canvas rotateBy: 45.
	canvas paintImage: image in: (0.4@0.4 corner: 0.6@0.6).! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 18:43'!
exampleImagePattern
	"self exampleImagePattern"
	| canvas image fill |
	canvas := self newExampleCanvas.
	image := self newExampleImage.
	fill := canvas newBitmapFill.
	fill image: image.
	fill repeatMode: #repeat.
	fill transform: (CMatrixTransform withScale: image extent * 5.0).
	canvas selectFill: fill.
	canvas drawRectangle: (0@0 corner: 1@1).
! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleLineCap
	"self exampleLineCap"
	| canvas |
	canvas := self newExampleCanvas.

	#(butt round square) do:[:capStyle|
		canvas selectPenWidth: 0.12 color: Color black.
		canvas capStyle: capStyle.
		canvas drawLineFrom: 0.25@0.2 to: 0.25@0.8.

		canvas selectPenWidth: 0.01 color: (Color r: 1.0 g: 0.2 b: 0.2).
		canvas capStyle: #round.
		canvas drawLineFrom: 0.25@0.2 to: 0.25@0.8.

		canvas translateBy: 0.25@0.0.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/19/2005 00:44'!
exampleLineJoin
	"self exampleLineJoin"
	| canvas |
	canvas := self newExampleCanvas.
	#(miter bevel round) do:[:joinStyle|
		canvas joinStyle: joinStyle.
		canvas drawPath:[
			canvas moveTo: 0.3@0.33.
			canvas lineBy: 0.2@-0.2.
			canvas lineBy: 0.2@0.2.
		].
		canvas translateBy: 0.0@0.3.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/17/2005 18:44'!
examplePath
	"self examplePath"
	| canvas |
	canvas := self newExampleCanvas.
	canvas drawPath:[
		canvas moveTo: 0.5@0.1.
		canvas lineTo: 0.9@0.9.
		canvas lineBy: -0.4@0.0.
		canvas curveVia: 0.2@0.9 and: 0.2@0.5 to: 0.5@0.5.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:50'!
exampleText
	"self exampleText"
	| canvas |
	canvas := self newExampleCanvas.
	canvas selectFont: 'Sans' size: 0.35 flags: 1. "bold"
	canvas drawString: 'Hello' at: 0.04@0.53.
	canvas selectFill: (Color r: 0.5 g: 0.5 b: 1.0).
	canvas selectPenWidth: 0.01 color: Color black.
	canvas drawPath:[
		canvas moveTo: 0.27@0.65.
		canvas text: 'void'.
	].! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:00'!
imageData
	^'/9j/4AAQSkZJRgABAQEASABIAAD/2wBDABALDA4MChAODQ4SERATGCgaGBYWGDEjJR0oOjM9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'! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:57'!
newExampleAlphaImage
	| image |
	image := self newExampleImage.
	image := image asFormOfDepth: 32.
	image fixAlpha.
	(BitBlt toForm: image)
		fillColor: (Color white alpha: 0.5);
		combinationRule: Form and;
		copyBits.
	^image! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'bf 7/31/2006 16:51'!
newExampleCanvas
	| canvas |
	canvas := self new.
	canvas scaleBy: (Display width min: Display height) asPoint.
	canvas selectPenWidth: 0.03 color: Color black.
	^canvas! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 19:55'!
newExampleImage
	| data |
	data := Base64MimeConverter mimeDecode: self imageData as: ByteString.
	^Form fromBinaryStream: data asByteArray readStream! !

!RomeReferenceCanvas class methodsFor: 'examples' stamp: 'ar 4/18/2005 20:00'!
newExamplePaintImage
	| image |
	image := self newExampleImage asFormOfDepth: 8.
	(Form dotOfSize: image width) displayOn: image at: 0@0 rule: Form and.
	^image! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'ar 4/16/2005 22:58'!
abort
	"Aborts the entire drawing process. This is used by printers or file output."! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'bf 5/18/2006 11:53'!
arc: radius at: aPoint from: angle1 to: angle2
	"Make an arc segment. The arc will be drawn in CW direction. 
	To reverse the direction of the arc use a negative radius. See
	Michael Goldapp, 'Approximation of circular arcs by cubic poynomials', 
	Computer Aided Geometric Design 8 (1991) 227-238."
	| delta h angleA angleB rAcos rAsin rBcos rBsin cp1 cp2 cp3 cp4 count step r  |
	radius negative
		ifFalse: [
			r := radius.
			delta := (angle2 - angle1) asFloat.
			angleA := angle1 degreesToRadians
		]
		ifTrue: [
			r := radius abs.
			delta := (angle1 - angle2) asFloat.
			angleA := angle2 degreesToRadians.
		].

	"FIXME: Compute better estimate for number of segments needed."
	delta < 0.0 ifTrue:[delta := delta + 360].
	count := (delta / 45.0) truncated max: 1.
	step := (delta / count asFloat) degreesToRadians.

	"The formula for computing the spline is
		(R * cos(A), R * sin(A))
		(R * cos(A) - h * sin(A), R * sin(A) + h * cos (A))
		(R * cos(B) + h * sin(B), R * sin(B) - h * cos (B))
		(R * cos(B), R * sin(B))

	for some value of h - the above paper gives us 

		h = 4/3 * tan(angle/4)

	as a very practical value"

	h := 4.0/3.0 * (step / 4.0) tan.

	0 to: count-1 do:[:i|
		angleB := angleA + step.
		rAcos := r * angleA cos.
		rAsin := r * angleA sin.
		rBcos := r * angleB cos.
		rBsin := r * angleB sin.
		cp1 := rAcos @ rAsin.
		cp2 := (rAcos - (h * rAsin)) @ (rAsin + (h * rAcos)).
		cp3 := (rBcos + (h * rBsin)) @ (rBsin - (h * rBcos)).
		cp4 := rBcos @ rBsin.

		(i = 0 and: [self currentPoint isNil]) ifTrue:[self moveTo: cp1 + aPoint].
		self curveVia: cp2 + aPoint and: cp3 + aPoint to: cp4 + aPoint.
		angleA := angleB.
	].! !

!RomeReferenceCanvas methodsFor: 'drawing images' stamp: 'bf 7/28/2006 18:28'!
blendImage: aForm in: destRect sourceRect: aRectangle
	"Draw a translucent image with the set of parameters"
	^self drawImage: aForm in: destRect sourceRect: aRectangle rule: 34! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/16/2005 23:06'!
capStyle
	"Answer the cap style for lines."
	^capStyle! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/17/2005 00:49'!
capStyle: aSymbol
	"Set the cap style for lines, either 
	#butt, #square, or #round"
	capStyle := aSymbol.
	self strokeChanged.! !

!RomeReferenceCanvas methodsFor: 'drawing' stamp: 'bf 7/29/2006 01:49'!
clear
	"clear current clip region to selected fill"
	self flag: #fixMe.
	"use actual clipping, like #clearPath: or something"
	self deviceClear: self deviceClip bounds fill: self deviceFill
! !

!RomeReferenceCanvas methodsFor: 'path functions' stamp: 'jl 7/12/2006 16:37'!
clipPath: aPathOrBlock
	"Construct a path and merge it with the current clip shape."
	| thePath devicePath clip |
	thePath := self makePath: aPathOrBlock.
	thePath isEmpty ifTrue:[
		^self deviceClip: (RomeClipshape new bounds: (0@0 corner: 0@0)).
	].
	"Transform path to device space"
	devicePath := thePath transformedBy: self transform.
	"Mix with clip path"
	clip := self deviceClip.
	(devicePath flatten: self deviceTolerance) do:[:poly|
		clip := clip intersect: (RomeClipshape new points: poly).
	].
	self deviceClip: clip.! !

!RomeReferenceCanvas methodsFor: 'path functions' stamp: 'bf 7/17/2006 14:11'!
clipRectangle: aRect

	((self transform isNoTransformWithin: aRect) not or: [deviceClip isRectangular not])
		ifTrue: [^self clipPath: [self rectangle: aRect]].

	^self deviceClip: (self deviceClip intersect: 
		(RomeClipshape new bounds: (aRect transformedBy: self transform)))! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'ar 4/16/2005 22:58'!
close
	"Closes any associated resources (if necessary).
	Closing a canvas also implies to #finish it."
	self finish.! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:52'!
closePath
	"Close the current path"
	path closePath.! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'bf 6/6/2006 14:03'!
createPath
	^RomePath new! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'bf 5/18/2006 11:53'!
currentPoint
	"If a path is under construction, answer its last point, nil otherwise"
	^path currentPoint! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:52'!
curveVia: cp1 and: cp2 to: aPoint
	"Add a curve to the current path"
	path curveVia: cp1 and: cp2 to: aPoint! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:52'!
curveVia: offCurve to: aPoint
	"Add a curve to the current path"
	path curveVia: offCurve to: aPoint.! !

!RomeReferenceCanvas methodsFor: 'path functions' stamp: 'ar 4/19/2005 00:54'!
dashPath: aPath
	"Construct a path and dash it according to dash spec"
	| loop lastPt dashes tolerance thePath pts |
	thePath := self makePath: aPath.
	"Turn off dashing internally"
	dashes := dashSpec.
	dashSpec := nil.
	loop := WriteStream on: (Array new: 10).
	tolerance := (self invertDistance: self deviceTolerance sqrt) squared.
	(thePath flatten: tolerance) do:[:poly|
		lastPt := nil.
		self dashPolyline: poly using: dashes do:[:pt1 :pt2|
			pt1 = lastPt ifTrue:[
				loop nextPut: pt2.
			] ifFalse:[
				(pts := loop contents) size = 0 ifFalse:[self drawPolyline: pts].
				loop resetToStart.
				loop nextPut: pt1; nextPut: pt2.
			].
			lastPt := pt2.
		].
		(pts := loop contents) size = 0 ifFalse:[self drawPolyline: pts].
		loop resetToStart.
	].
	dashSpec := dashes.! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'ar 4/17/2005 23:00'!
dashPolyline: points using: dashes do: aBlock
	"Dash the given polyline."
	| dash dashMax dashIndex dashOn lastPt nextPt offset line 
	remainder lineMax lineIndex delta p1 p2 closed |
	closed := points first = points last.
	dashMax := dashes size.
	dashMax = 0 ifTrue:[^self].
	lineMax := points size.
	lineMax < 2 ifTrue:[^self].
	dashIndex := 1.
	dashOn := true.
	dash := dashes at: 1.

	dashOffset > 0.0 ifTrue:[
		"Advance by dashOffset"
		offset := dashOffset.
		[offset > dash] whileTrue:[
			offset := offset - dash.
			(dashIndex := dashIndex + 1) > dashMax ifTrue:[dashIndex := 1].
			dashOn := dashOn not.
		].
	].


	"Now dash the poly line"
	closed ifTrue:[
		p1 := lastPt := points at: lineMax.
		nextPt := points at: (lineIndex := 1).
	] ifFalse:[
		p1 := lastPt := points at: 1.
		nextPt := points at: (lineIndex := 2).
	].

	delta := nextPt - lastPt.
	line := remainder := delta length.
	[true] whileTrue:[
		remainder <= dash ifTrue:[
			"The line is fully inside this dash"
			dashOn ifTrue:[aBlock value: p1 value: nextPt].
			dash := dash - remainder.
			"Advance to next segment"
			(lineIndex := lineIndex + 1) > lineMax ifTrue:[^self].
			p1 := lastPt := nextPt.
			nextPt := points at: lineIndex.
			delta := nextPt - lastPt.
			line := remainder := delta length.
		] ifFalse:[
			"The line segment is longer than the dash"
			p2 := p1 + (delta * dash / line).
			remainder := remainder - dash.
			dashOn ifTrue:[aBlock value: p1 value: p2].
			p1 := p2.
			"Advance to next dash"
			(dashIndex := dashIndex + 1) > dashMax ifTrue:[dashIndex := 1].
			dash := dashes at: dashIndex.
			dashOn := dashOn not.
		].
	].
! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/16/2005 23:07'!
dashSpec
	"Answer the current dash specification"
	^dashSpec ifNil:[#()].! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/17/2005 00:49'!
dashSpec: aSpec
	"Set the new dash specification"
	dashSpec := aSpec.
	dashSpec size = 0 ifTrue:[dashSpec := nil].
	self strokeChanged.! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'bf 7/29/2006 01:45'!
deviceClear: aRectangle fill: aFill
	"set aRectangle to aFill *even* if aFill is transparent."
	Display 	getCanvas
		clipBy: aRectangle during: [:c | c fillColor: aFill]! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'ar 4/19/2005 00:55'!
deviceClip
	"Answer the currently selected clip shape"
	^deviceClip! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'ar 4/19/2005 00:55'!
deviceClip: aClipshape
	"Set the currently active clip shape"
	deviceClip := aClipshape.! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'ar 4/19/2005 00:56'!
deviceFill
	"Answer the current device fill. Usually this should be cached."
	^deviceFill! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'ar 4/19/2005 00:56'!
deviceFill: aFill
	"Set the current device fill."
	deviceFill := aFill.! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'bf 5/16/2006 12:56'!
deviceGeneralPolygon: aCollection pen: aPen fill: aFill
	"Send a general polygon (e.g., a polygon that can contain holes) directly to the device. If this is not supported, fall back to simple polygons"
	aCollection isEmpty ifTrue: [^nil].
	aCollection size = 1 ifTrue: [^self devicePolygon: aCollection first pen: aPen fill: aFill].

	"fill first"
	aFill = noFill ifFalse: [
		(self tesselate: aCollection) do: [:poly | 
			self devicePolygon: poly pen: noPen fill: aFill]].

	"draw contours separately"
	aPen = noPen ifFalse: [
		aCollection do: [:poly |
			self devicePolygon: poly pen: aPen fill: noFill]].
! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'jl 5/31/2006 12:28'!
deviceImage: aForm in: destRect sourceRect: sourceRect rule: rule
	"Draw the image; don't transform."
	| bb |
	destRect extent = sourceRect extent ifTrue:[
		"fine - just a plain blt"
		bb := BitBlt toForm: Display.
		bb colorMap: (aForm colormapIfNeededFor: Display).
		bb sourceForm: aForm.
		bb sourceOrigin: sourceRect origin.
		bb destOrigin: destRect origin.
"		bb extent: sourceRect extent."
		bb combinationRule: rule.
		bb copyBits.
	] ifFalse:[
		"not a plain blt - warp it"
		bb := WarpBlt toForm: Display.
		bb colorMap: (aForm colormapIfNeededFor: Display).
		bb sourceForm: aForm.
		bb sourceQuad: sourceRect innerCorners destRect: destRect.
		bb combinationRule: rule.
		bb warpBits.
	].
! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'ar 4/18/2005 22:29'!
deviceLineFrom: startPoint to: endPoint pen: aPen
	"Send a line directly to the device"
	aPen isSolid ifFalse:[
		self saveState.
		self transform: CMatrixTransform identity.
		aPen drawLineFrom: startPoint to: endPoint on: self.
		^self restoreState.
	].
	(BalloonCanvas on: Display)
		aaLevel: 1;
		line: startPoint to: endPoint width: aPen width color: aPen color.! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'ar 4/20/2005 01:23'!
deviceOval: outerBounds pen: aPen fill: aFill
	"Send an oval directly to the device. If a device does not support ovals it might just generalize into a path (such as shown below). If the device *does* support ovals but not the choosen pen and fill, it might ask the pen and fill to draw an oval by itself. In this case the device needs to provide a *new* canvas to be used by pen and fill so that the receiver is not affected by anything that pen and fill are doing."
	| rect |
	rect := aPen insetBounds: outerBounds.
	(aPen isSolid and:[aFill isSolidFill]) ifFalse:[
		"Draw pen and fill separately"
		self saveState.
		self transform: CIdentityTransform new.
		aFill isSolidFill
			ifTrue:[self deviceOval: outerBounds pen: noPen fill: aFill]
			ifFalse:[aFill fillOval: rect on: self].
		aPen isSolid
			ifTrue:[self deviceOval: outerBounds pen: aPen fill: noFill]
			ifFalse:[aPen frameOval: rect on: self].
		^self restoreState.
	].
	self devicePath: (self makePath:[self oval: rect]) pen: aPen fill: aFill.! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'bf 5/16/2006 13:08'!
devicePath: aPath pen: aPen fill: aFill
	"Send a path directly to the device. If a device does not support paths it might fall back onto a polygon (such as shown below). If the device *does* support paths but not the choosen pen and fill, it might ask the pen and fill to draw the path by itself. In this case the device needs to provide a *new* canvas to be used by pen and fill so that the receiver is not affected by anything that pen and fill are doing."
	| closedContours openContours |
	(aPen isSolid and:[aFill isSolidFill]) ifFalse:[
		"Draw pen and fill separately"
		self saveState.
		self transform: CIdentityTransform new.
		aFill isSolidFill
			ifTrue:[self devicePath: aPath pen: noPen fill: aFill]
			ifFalse:[aFill fillPath: aPath on: self].
		aPen isSolid
			ifTrue:[self devicePath: aPath pen: aPen fill: noFill]
			ifFalse:[aPen framePath: aPath on: self].
		^self restoreState.
	].

	closedContours := OrderedCollection new.
	openContours := OrderedCollection new.
	(aPath flatten: self deviceTolerance) do: [:poly |
		(poly first = poly last)
			ifTrue: [closedContours add: poly]
			ifFalse: [openContours add: poly]	].

	self deviceGeneralPolygon: closedContours pen: aPen fill: aFill.
	openContours do: [:poly | self devicePolyline: poly pen: aPen].

! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'bf 7/19/2006 19:05'!
devicePen
	"Answer the current device pen"
	^pen transformedBy: self transform! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'ar 4/18/2005 22:29'!
devicePolygon: aCollection pen: aPen fill: aFill
	"Send a polygon directly to the device. All devices must support polygons.  However, if the device not support the choosen pen and fill, it might ask the pen and fill to draw the polygon by itself. In this case the device needs to provide a *new* canvas to be used by pen and fill so that the receiver is not affected by anything that pen and fill are doing."
	aCollection isEmpty ifTrue:[^nil].
	(aPen isSolid and:[aFill isSolidFill]) ifFalse:[
		"Draw pen and fill separately"
		self saveState.
		self transform: CIdentityTransform new.
		aFill isSolidFill
			ifTrue:[self devicePolygon: aCollection pen: noPen fill: aFill]
			ifFalse:[aFill fillPolygon: aCollection on: self].
		aPen isSolid
			ifTrue:[self devicePolygon: aCollection pen: aPen fill: noFill]
			ifFalse:[aPen framePolygon: aCollection on: self].
		^self restoreState.
	].
	(BalloonCanvas on: Display)
		aaLevel: 1;
		drawPolygon: aCollection
		color: aFill
		borderWidth: aPen width
		borderColor: aPen color.! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'ar 4/18/2005 22:29'!
devicePolyline: aCollection pen: aPen
	"Send a polyline directly to the device"
	| lastPt |
	(aPen isSolid) ifFalse:[
		"Draw pen and fill separately"
		self saveState.
		self transform: CIdentityTransform new.
		aPen drawPolyline: aCollection on: self.
		^self restoreState.
	].

	lastPt := nil.
	aCollection do:[:nextPt|
		lastPt ifNotNil:[self deviceLineFrom: lastPt to: nextPt pen: aPen].
		lastPt := nextPt.
	].! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'ar 4/18/2005 22:29'!
deviceRect: outerBounds pen: aPen fill: aFill
	"Send a rectangle directly to the device. If a device does not support rectangles (how dare it?) it might just generalize into a path (such as shown below). If the device *does* support rectangles but not the choosen pen and fill, it might ask the pen and fill to draw a rectangle by itself. In this case the device needs to provide a *new* canvas to be used by pen and fill so that the receiver is not affected by anything that pen and fill are doing."
	| rect |
	rect := aPen insetBounds: outerBounds.
	(aPen isSolid and:[aFill isSolidFill]) ifFalse:[
		"Draw pen and fill separately"
		self saveState.
		self transform: CIdentityTransform new.
		aFill isSolidFill
			ifTrue:[self deviceRect: outerBounds pen: noPen fill: aFill]
			ifFalse:[aFill fillRectangle: rect on: self].
		aPen isSolid
			ifTrue:[self deviceRect: outerBounds pen: aPen fill: noFill]
			ifFalse:[aPen frameRectangle: rect on: self].
		^self restoreState.
	].
	self devicePath: (self makePath:[self rectangle: rect]) pen: aPen fill: aFill.! !

!RomeReferenceCanvas methodsFor: 'device responsibilities' stamp: 'ar 4/18/2005 22:29'!
deviceRoundRect: outerBounds radius: radius corners: flags pen: aPen fill: aFill
	"Send a rectangle directly to the device. If a device does not support round rects it might just generalize into a path (such as shown below). If the device *does* support round rects but not the choosen pen and fill, it might ask the pen and fill to draw a round rect by itself. In this case the device needs to provide a *new* canvas to be used by pen and fill so that the receiver is not affected by anything that pen and fill are doing."
	| rect |
	rect := aPen insetBounds: outerBounds.
	(aPen isSolid and:[aFill isSolidFill]) ifFalse:[
		"Draw pen and fill separately"
		self saveState.
		self transform: CIdentityTransform new.
		aFill isSolidFill
			ifTrue:[self deviceRoundRect: outerBounds radius: radius corners: flags pen: noPen fill: aFill]
			ifFalse:[aFill fillRoundRect: rect radius: radius corners: flags on: self].
		aPen isSolid
			ifTrue:[self deviceRoundRect: outerBounds radius: radius corners: flags pen: aPen fill: noFill]
			ifFalse:[aPen frameRoundRect: rect radius: radius corners: flags on: self].
		^self restoreState.
	].
	self devicePath: (self makePath:[self roundRect: rect radius: radius corners: flags]) 
			pen: aPen fill: aFill.! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'ar 4/19/2005 00:56'!
deviceTolerance
	^deviceTolerance! !

!RomeReferenceCanvas methodsFor: 'device state' stamp: 'ar 4/19/2005 00:56'!
deviceTolerance: aNumber
	deviceTolerance := aNumber.! !

!RomeReferenceCanvas methodsFor: 'drawing curves' stamp: 'bf 7/24/2006 17:03'!
drawArc: radius at: aPoint from: angle1 to: angle2
	"Draw an arc segment. The arc will be drawn in CW direction. 
	To reverse the direction of the arc use a negative radius"
	^self drawPath:[self arc: radius at: aPoint from: angle1 to: angle2].! !

!RomeReferenceCanvas methodsFor: 'drawing curves' stamp: 'ar 4/19/2005 00:47'!
drawCurveFrom: start via: cp1 and: cp2 to: end
	"Draw a cubic bezier curve"
	^self drawPath:[
		self moveTo: start.
		self curveVia: cp1 and: cp2 to: end.
	].! !

!RomeReferenceCanvas methodsFor: 'drawing curves' stamp: 'ar 4/19/2005 00:47'!
drawCurveFrom: start via: offCurve to: end
	"Draw a quadratic bezier curve"
	^self drawPath:[
		self moveTo: start.
		self curveVia: offCurve to: end.
	].! !

!RomeReferenceCanvas methodsFor: 'drawing images' stamp: 'ar 4/18/2005 19:36'!
drawImage: aForm in: destRect sourceRect: sourceRect
	"Draw a fully opaque image with the set of parameters"
	^self drawImage: aForm in: destRect sourceRect: sourceRect rule: Form over! !

!RomeReferenceCanvas methodsFor: 'drawing images' stamp: 'bf 7/25/2006 14:16'!
drawImage: aForm in: aRectangle sourceRect: sourceRect rule: rule
	"Draw an image with the set of parameters"
	| rect clip pt1 pt2 bmFill |
	"Check for weirdo transformations"
	(ctm isNoTransformWithin: aRectangle) ifTrue:[
		"Check for weirdo clipping"
		pt1 := self transformPoint: aRectangle origin.
		pt2 := self transformPoint: aRectangle corner.
		rect := pt1 corner: pt2.
		((clip := self deviceClip) isRectangular or:[clip containsRect: rect]) ifTrue:[
			"Send image straight to device"
			^self deviceImage: aForm in: rect sourceRect: sourceRect rule: rule
		].
	].
	"Generalize image to fill and draw rectangle"
	bmFill := CBitmapFill form: aForm.
	bmFill tileFlag: false.
	bmFill origin: 0@0.
	bmFill direction: aForm width@0.
	bmFill normal: 0@aForm height.
	bmFill combinationRule: rule.
	self saveState.
		self translateBy: aRectangle origin.
		self scaleBy: aRectangle extent / sourceRect extent.
		self selectFill: bmFill.
		self selectNoPen.
		self drawRectangle: sourceRect.
	self restoreState.! !

!RomeReferenceCanvas methodsFor: 'drawing curves' stamp: 'ar 4/19/2005 00:47'!
drawLineFrom: startPoint to: endPoint
	"Draw a line from startPoint to endPoint."
	| pt1 pt2  rect clip |
	self devicePen width <= 0 ifTrue:[^self].

	pt1 := ctm transformPoint: startPoint.
	pt2 := ctm transformPoint: endPoint.
	rect := self devicePen expandBounds: (pt1 rect: pt2).
	((clip := self deviceClip) isRectangular or:[clip containsRect: rect]) ifFalse:[
		^self drawPath:[self line: startPoint to: endPoint].
	].
	"See if we have a weirdo stroke to deal with"
	(dashSpec == nil and:[capStyle == #round and:[joinStyle == #round]]) ifFalse:[
		^self drawPath:[self line: startPoint to: endPoint].
	].
	^self deviceLineFrom: pt1 to: pt2 pen: self devicePen! !

!RomeReferenceCanvas methodsFor: 'drawing shapes' stamp: 'bf 5/18/2006 13:42'!
drawOval: aRectangle
	"Draw an oval specified by the given rectangle"
	| pt1 pt2  rect clip |
	"Check for weirdo transformations"
	(ctm isNoTransformWithin: aRectangle) ifFalse:[
		^self drawPath:[self oval: aRectangle].
	].
	"Check for weirdo clipping"
	pt1 := self transformPoint: aRectangle origin.
	pt2 := self transformPoint: aRectangle corner.
	rect := self devicePen expandBounds: ((pt1 min: pt2) corner: (pt1 max: pt2)).
	((clip := self deviceClip) isRectangular or:[clip containsRect: rect]) ifFalse:[
		^self drawPath:[self oval: aRectangle]
	].
	"Check for weirdo line attributes"
	(dashSpec == nil) ifFalse:[
		self deviceOval: rect pen: noPen fill: self deviceFill.
		self saveState.
			self selectFill: noFill.
			self drawPath:[self oval: aRectangle].
		^self restoreState.
	].
	"Send geometry straight to device"
	self deviceOval: rect pen: self devicePen fill: self deviceFill! !

!RomeReferenceCanvas methodsFor: 'drawing shapes' stamp: 'bf 7/26/2006 17:35'!
drawPath: aBlock
	"Draw a generic path."
	| devicePath weirdoStroke rect clip aPath devicePoly |
	aPath := self makePath: aBlock.
	aPath isEmpty ifTrue:[^self].

	"Transform path to device space"
	devicePath := aPath transformedBy: self transform.

	"See if we have a weirdo stroke to deal with"
	weirdoStroke := false.
	(dashSpec == nil and:[capStyle == #round and:[joinStyle == #round]]) ifFalse:[
		weirdoStroke := self devicePen width > 0.
	].

	"See if we need to clip - in this case we must compute the outline first, 
	since clip might clip parts of the outline."
	rect := nil. "lazy"
	((clip := self deviceClip) isRectangular or: [
		rect := self devicePen expandBounds: devicePath bounds.
		clip containsRect: rect]) ifTrue:[
			"All right, we can leave the clipping to device."
			(weirdoStroke) ifFalse:[
				^self devicePath: devicePath pen: self devicePen fill: self deviceFill
			].
			"Only do fill - weirdo stroke will be handled below"
			self devicePath: devicePath pen: noPen fill: self deviceFill.
		] ifFalse:[
			"Device cannot clip; do it ourselves"
			rect ifNil: [rect := self devicePen expandBounds: devicePath bounds].
			(self deviceClip bounds intersects: rect) ifFalse:[^self]. "quick reject"
			aPath closed ifTrue: [
				"Fill interior first"
				devicePoly := devicePath flatten: self deviceTolerance.
				devicePoly := self deviceClip intersectGeneralPoly: devicePoly.
				self deviceGeneralPolygon: devicePoly pen: noPen fill: self deviceFill
			].
		].
	"Now stroke path"
	self devicePen width <= 0 ifTrue:[^self].

	dashSpec ifNotNil:[^self dashPath: aPath].

	(devicePath flatten: self deviceTolerance) do:[:poly|
		| penPolys fill |
		fill := self devicePen fill.
		penPolys := self tracePolygon: poly.
		penPolys := self deviceClip intersectGeneralPoly: penPolys.
		penPolys do: [:pts | self devicePolygon: pts pen: noPen fill: fill].
		"if penPolys would not be overlapping, we could use this:"
		"self deviceGeneralPolygon: penPoly pen: noPen fill: self devicePen fill."
	].! !

!RomeReferenceCanvas methodsFor: 'drawing shapes' stamp: 'ar 4/17/2005 18:48'!
drawPolygon: aCollection
	"Draw a polygon using the given set of vertices"
	^self drawPath:[self polygon: aCollection].! !

!RomeReferenceCanvas methodsFor: 'drawing curves' stamp: 'ar 4/19/2005 00:47'!
drawPolyline: aCollection
	"Draw a polyline using the given set of vertices"
	^self drawPath:[self polyline: aCollection].! !

!RomeReferenceCanvas methodsFor: 'drawing shapes' stamp: 'bf 6/8/2006 14:45'!
drawRectangle: aRectangle
	"Draw aRectangle."
	| pt1 pt2  rect clip |
	"Check for weirdo transformations"
	(ctm isNoTransformWithin: aRectangle) ifFalse:[
		^self drawPath:[self rectangle: aRectangle].
	].
	"Check for weirdo clipping"
	pt1 := self transformPoint: aRectangle origin.
	pt2 := self transformPoint: aRectangle corner.
	rect := self devicePen expandBounds: ((pt1 min: pt2) corner: (pt1 max: pt2)).
	((clip := self deviceClip) isRectangular or:[clip containsRect: rect]) ifFalse:[
		^self drawPath:[self rectangle: aRectangle].
	].

	"Check for weirdo line attributes"
	(dashSpec == nil and:[joinStyle == #miter or: [self devicePen hasEffect not]]) ifFalse:[
		self deviceRect: rect pen: noPen fill: self deviceFill.
		self saveState.
			self selectFill: noFill.
			self drawPath:[self rectangle: aRectangle].
		^self restoreState.
	].

	"Send geometry straight to device"
	self deviceRect: rect pen: self devicePen fill: self deviceFill! !

!RomeReferenceCanvas methodsFor: 'drawing shapes' stamp: 'bf 5/18/2006 13:38'!
drawRoundRect: aRectangle radius: aRadius corners: flags
	"Draw a round rectangle using the specified position 
	and size, radius, and corners to round."
	| pt1 pt2  rect radius clip |
	"Check for weirdo transformations"
	(ctm isNoTransformWithin: aRectangle) ifFalse:[
		^self drawPath:[self roundRect: aRectangle radius: aRadius corners: flags]
	].
	"Check for weirdo clipping"
	pt1 := self transformPoint: aRectangle origin.
	pt2 := self transformPoint: aRectangle corner.
	rect := self devicePen expandBounds: ((pt1 min: pt2) corner: (pt1 max: pt2)).
	((clip := self deviceClip) isRectangular or:[clip containsRect: rect]) ifFalse:[
		^self drawPath:[self roundRect: aRectangle radius: aRadius corners: flags]
	].

	radius := self transformDistance: aRadius.

	"Check for weirdo line attributes"
	(dashSpec == nil) ifFalse:[
		self deviceRoundRect: rect radius: radius corners: flags pen: noPen fill: self deviceFill.
		self saveState.
			self selectFill: noFill.
			self drawPath:[self roundRect: aRectangle radius: aRadius corners: flags].
		self restoreState.
	].

	"Send geometry straight to device"
	self deviceRoundRect: rect radius: radius corners: flags 
			pen: self devicePen fill: self deviceFill! !

!RomeReferenceCanvas methodsFor: 'drawing text' stamp: 'bf 7/31/2006 15:53'!
drawString: aString at: aPoint from: start to: end
	^font render: aString at: aPoint from: start to: end on: self! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'ar 4/17/2005 14:50'!
eigenValuesOf: aTransform
	"Compute the eigenvalues of aTransform."
	| m a b c d radius |
	m := aTransform asMatrixTransform2x3.
	a := m a11.	b := m a12.
	c := m a21.	d := m a22.
	radius := ((a*a) + (2.0*a*d) + (d*d) - (4.0*a*d) - (4.0*b*c)) sqrt.
	^Array
		with: (a + d + radius) * 0.5
		with: (a + d - radius) * 0.5! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'ar 4/16/2005 22:45'!
finish
	"Flushes the canvas and ensures that he effects are
	visible on the underlying output medium."! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'ar 4/16/2005 22:58'!
flush
	"Flushes the canvas, e.g., ensures that no state is 
	buffered in the canvas any longer."
	self finish. "if #flush isn't implemented but #finish then use #finish"! !

!RomeReferenceCanvas methodsFor: 'attributes text' stamp: 'ar 4/16/2005 23:19'!
font
	"Answer the currently selected font"
	^font! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'ar 4/16/2005 23:21'!
fontChanged
	"The current font has changed. Do whatever is necessary."! !

!RomeReferenceCanvas methodsFor: 'attributes text' stamp: 'ar 4/16/2005 23:20'!
fontMatrix
	"Answer the current transform for fonts"
	^fontMatrix! !

!RomeReferenceCanvas methodsFor: 'attributes text' stamp: 'ar 4/16/2005 23:20'!
fontMatrix: aTransform
	"Set the matrix transforming from font design units to user space.
	Normally this is just a scale (via fontSize) but can be overridden
	to achieve shearing and other effects."
	fontMatrix := aTransform.! !

!RomeReferenceCanvas methodsFor: 'attributes text' stamp: 'jl 6/7/2006 17:48'!
fontSize
	"Answer the currently selected font size."
	^fontMatrix scale y! !

!RomeReferenceCanvas methodsFor: 'attributes text' stamp: 'jl 6/7/2006 17:49'!
fontSize: aNumber
	"Select a new font size for the active font."
	fontMatrix := CMatrixTransform withScale: aNumber! !

!RomeReferenceCanvas methodsFor: 'attributes text' stamp: 'ar 4/16/2005 23:20'!
font: aFont
	"Set the font to be used for subsequent drawing"
	font := aFont.
	self fontChanged.! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'bf 7/19/2006 19:07'!
initialize
	"Initializes the canvas with default values"
	noPen := RomePen width: 0 color: Color transparent.
	noFill := Color transparent.
	noClip := RomeClipshape new bounds: (0@0 corner: 1000@1000).

	"Basic attributes"
	ctm := CIdentityTransform new.
	deviceFill := noFill.
	pen := noPen.
	deviceClip := noClip.
	deviceTolerance := 0.5 squared.

	"Line attributes"
	capStyle := #round.
	joinStyle := #round.
	miterLimit := 10.
	dashSpec := nil.
	dashOffset := 0.0.


	"Font attributes"
	font := nil.
	self fontSize: 1.
	fontMatrix := CMatrixTransform identity.! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/17/2005 22:29'!
invertDistance: aNumber
	"Transform the given distance from device to user space"
	"@@@@: FIXME - this is incorrect."
	^(ctm inverseTransformation transformDirection: aNumber@0) r! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/16/2005 23:10'!
invertPoint: aPoint
	"Transform the given point from device to user space"
	^self transform invertPoint: aPoint! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/16/2005 23:07'!
joinStyle
	"Answer the join style for lines"
	^joinStyle! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/17/2005 00:49'!
joinStyle: aSymbol
	"Set the join style for lines, either
	#miter, #round, or #bevel"
	joinStyle := aSymbol.
	self strokeChanged.! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:52'!
lineBy: aPoint
	path lineBy: aPoint.! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:52'!
lineTo: aPoint
	path lineTo: aPoint.! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:52'!
line: startPoint to: endPoint
	self moveTo: startPoint.
	self lineTo: endPoint.! !

!RomeReferenceCanvas methodsFor: 'path functions' stamp: 'bf 6/6/2006 14:03'!
makePath: aBlockOrPath
	"Construct a path from either path or block"
	| donePath |
	path ifNotNil:[self error: 'Old path still active'].
	aBlockOrPath isBlock ifFalse:[^aBlockOrPath].
	path := self createPath.
	aBlockOrPath value.
	donePath := path.
	path := nil.
	^donePath! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/16/2005 23:07'!
miterLimit
	"Answer the current miter limit"
	^miterLimit! !

!RomeReferenceCanvas methodsFor: 'attributes lines' stamp: 'ar 4/17/2005 00:49'!
miterLimit: aNumber
	"Indicate the current miter limit"
	miterLimit := aNumber.
	self strokeChanged.! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:53'!
moveTo: aPoint
	path moveTo: aPoint.
! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'ar 4/16/2005 22:45'!
open
	"Opens any associated resources (if necessary."! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:53'!
oval: aRectangle
	"Convert an oval into a path"
	| width height angleStep h angleA angleB rAcos rAsin rBcos rBsin 
	cp1 cp2 cp3 cp4 radius offset scale |
	offset := aRectangle origin + aRectangle corner * 0.5.
	width := aRectangle width * 0.5.
	height := aRectangle height * 0.5.
	width >= height ifTrue:[
		radius := width.
		scale := 1.0 @ (height / width).
	] ifFalse:[
		radius := height.
		scale := (width / height) @ 1.0.
	].
	"FIXME: Choose better approximation"
	angleStep := 45.0 degreesToRadians.
	h := 4.0/3.0 * (angleStep / 4.0) tan.
	0 to: 7 do:[:i|
		angleA := i * angleStep.
		angleB := angleA + angleStep.

		rAcos := radius * angleA cos.
		rAsin := radius * angleA sin.
		rBcos := radius * angleB cos.
		rBsin := radius * angleB sin.

		cp1 := rAcos @ rAsin.
		cp2 := (rAcos - (h * rAsin)) @ (rAsin + (h * rAcos)).
		cp3 := (rBcos + (h * rBsin)) @ (rBsin - (h * rBcos)).
		cp4 := rBcos @ rBsin.

		i = 0 ifTrue:[self moveTo: cp1*scale+offset].
		self curveVia: cp2*scale+offset and: cp3*scale+offset to: cp4*scale+offset.
	].
	self closePath.! !

!RomeReferenceCanvas methodsFor: 'drawing images' stamp: 'ar 4/18/2005 19:38'!
paintImage: aForm in: destRect sourceRect: aRectangle
	"Draw a color keyed image with the set of parameters"
	^self drawImage: aForm in: destRect sourceRect: aRectangle rule: Form paint! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'jl 7/4/2006 16:23'!
path: aPath
	"add aPath to current path"

	aPath pathOn: self! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:53'!
polygon: aCollection
	"Make a polygon"
	self polyline: aCollection.
	self closePath.! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:53'!
polyline: aCollection
	"Make a polyline"
	self moveTo: aCollection first.
	2 to: aCollection size do:[:i| self lineTo: (aCollection at: i)].! !

!RomeReferenceCanvas methodsFor: 'preserve state' stamp: 'bf 7/14/2006 14:30'!
preserveStateDuring: aBlock
	self saveState.
	^aBlock ensure: [	self restoreState].
! !

!RomeReferenceCanvas methodsFor: 'preserve state' stamp: 'bf 7/14/2006 14:30'!
preserveTransformDuring: aBlock
	"Reset transform after evaluating aBlock"
	| oldTransform |
	oldTransform := self transform.
	^aBlock ensure: [self transform: oldTransform].! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'ar 4/19/2005 00:53'!
rectangle: rect
	"Make a rectangle"
	self moveTo: rect topLeft.
	self lineTo: rect topRight.
	self lineTo: rect bottomRight.
	self lineTo: rect bottomLeft.
	self closePath.! !

!RomeReferenceCanvas methodsFor: 'preserve state' stamp: 'ar 4/17/2005 03:06'!
restoreState
	"Restore the state from the internal stack"
	self copyFrom: stack.! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/16/2005 23:08'!
rotateBy: angle
	"Rotate all further drawing operations by angle"
	^self transformBy: (CMatrixTransform withAngle: angle)! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'bf 7/31/2006 14:10'!
roundRect: rect radius: radius corners: cornerFlags
	"Convert a round rect into a path"
	| outer inner r x y |
	x := rect width / 2.0.
	y := rect height / 2.0.
	r := ((radius min: x) min: y).

	outer := rect.
	inner := outer insetBy: r.
	(cornerFlags anyMask: 1) ifTrue:[
		self moveTo: outer left @ inner top.
		self arc: r at: inner topLeft from: 180 to: 270.
	] ifFalse:[
		self moveTo: outer topLeft.
	].

	(cornerFlags anyMask: 8) ifTrue:[
		self lineTo: inner right @ outer top.
		self arc: r at: inner topRight from: 270 to: 360.
	] ifFalse:[
		self lineTo: outer topRight.
	].

	(cornerFlags anyMask: 4) ifTrue:[
		self lineTo: outer right @ inner bottom.
		self arc: r at: inner bottomRight from: 0 to: 90.
	] ifFalse:[
		self lineTo: outer bottomRight.
	].

	(cornerFlags anyMask: 2) ifTrue:[
		self lineTo: inner left @ outer bottom.
		self arc: r at: inner bottomLeft from: 90 to: 180.
	] ifFalse:[
		self lineTo: outer bottomLeft.
	].

	self closePath.
! !

!RomeReferenceCanvas methodsFor: 'preserve state' stamp: 'ar 4/17/2005 03:05'!
saveState
	"Save the state of the canvas onto the internal stack."
	stack := self clone.
! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/16/2005 23:08'!
scaleBy: scale
	"Scale all further drawing operations by scale"
	^self transformBy: (CMatrixTransform withScale: scale)
! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'ar 4/17/2005 14:13'!
selectClip: aClipshape
	"Clip the receiver by aClipshape, which can be either
	a Clipshape, a Rectangle, a Path, or a Mask."
	| globalClip |
	globalClip := aClipshape transformedBy: self transform.
	^self deviceClip: (self deviceClip intersect: globalClip)! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'ar 4/17/2005 20:07'!
selectFill: aFill
	"Select the fill to use to use for subsequent drawing.
	This is a write-only operation."
	| globalFill |
	globalFill := aFill isColor ifTrue:[aFill] ifFalse:[aFill transformedBy: self transform].
	self deviceFill: globalFill.! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'jl 6/21/2006 10:16'!
selectNoFill
	deviceFill := noFill! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'bf 7/19/2006 19:06'!
selectNoPen
	pen := noPen! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'bf 7/31/2006 17:08'!
selectPenWidth: aNumber color: aColor
	"select a solid Pen of width and color"
	DefaultPen ifNil: [DefaultPen := RomePen new].
	DefaultPen width: aNumber color: aColor.
	self selectPen: DefaultPen! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'bf 7/19/2006 19:06'!
selectPen: aPen
	"Set the pen to use for subsequent drawing."
	pen := aPen.
	! !

!RomeReferenceCanvas methodsFor: 'attributes canvas ' stamp: 'ar 4/17/2005 23:10'!
selectTolerance: aNumber
	"Indicate the tolerance for converting paths.
	The tolerance is the distance by which a flattened
	path might deviate from its original version."
	^self deviceTolerance: (self transformDistance: aNumber) squared! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'bf 7/24/2006 16:34'!
setTarget: aForm
! !

!RomeReferenceCanvas methodsFor: 'initialize-release' stamp: 'ar 4/16/2005 22:45'!
showPage
	"Printers only. Advances to the next page of output."! !

!RomeReferenceCanvas methodsFor: 'drawing images' stamp: 'bf 7/29/2006 00:00'!
stencilImage: aForm in: destRect sourceRect: aRectangle
	"Fill destRect using aForm's alpha channel as mask"
	
	| mask |
	mask := Form extent: aRectangle extent depth: 32.
	mask getCanvas stencil: aForm at: 0@0 sourceRect: aRectangle color: deviceFill.
	^self blendImage: mask in: destRect! !

!RomeReferenceCanvas methodsFor: 'drawing text' stamp: 'bf 7/31/2006 16:27'!
strikeString: aString at: aPoint from: startIndex to: stopIndex
	"fill has been set, now just draw the string."
	| pos glyphs xTable height glyphIndex leftX rightX |
	pos := aPoint - (0@font asStrikeFont ascent) .
	glyphs := font asStrikeFont glyphs.
	xTable := font asStrikeFont xTable.
	height := glyphs height.
	startIndex to: stopIndex do: [:i |
		glyphIndex := (aString at: i) charCode + 1.
		leftX := xTable at: glyphIndex.
		rightX := xTable at: glyphIndex + 1.
		self stencilImage: glyphs
			at: pos
			sourceRect: (leftX @ 0 corner: rightX @ height).
		pos := pos + (rightX - leftX @ 0).
	]
! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'bf 5/16/2006 15:50'!
string: aString
	
	self string: aString from: 1 to: aString size! !

!RomeReferenceCanvas methodsFor: 'path creation' stamp: 'bf 7/14/2006 14:43'!
string: aString from: startIndex to: stopIndex

	| offset advance |
	offset := 0@0.
	startIndex to: stopIndex do: [:index |
		advance := font outlineChar: (aString at: index)
			size: 1
			transform: (fontMatrix composedWithLocal: (COffsetTransform offset: offset))
			on: self.
		offset := offset + advance].
	^offset! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'ar 4/17/2005 00:50'!
strokeChanged
	"The current stroke attributes have changed. Do whatever is necessary."! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'bf 5/16/2006 13:00'!
tesselate: aCollection
	"Tesselate a general polygon into a collection of simple polygons"

	self flag: #fixMe.
	Transcript cr; show: 'Rome: tesselation not supported yet'.
	^aCollection! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'bf 7/27/2006 17:09'!
tracePolygon: inPoints
	"Compute the minkowski sum of the shape of the devicePen with the polygon/polyline specified by the argument."

	| r a b d n |
	self flag: #fixMe. "neither correct nor optimized, does not do capstyle/joinstyle"
	r := self devicePen width / 2.0.
	^Array streamContents: [:outContours |
		1 to: inPoints size-1 do: [:i |
			a := inPoints at: i.
			b := inPoints at: i+1.
			a = b ifFalse: [
				d := (b - a) normalized * r.		"direction"
				n := d y negated @ d x.			"normal"
				outContours nextPut: {a-n. a-d. a+n. b+n. b+d. b-n}.
			].
		].
	].! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/17/2005 14:02'!
transform
	"Answer the currently selected transform"
	^ctm! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/16/2005 23:08'!
transformBy: aTransform
	"Transform all further drawing operations by aTransform"
	^self transform: (self transform composedWithLocal: aTransform)! !

!RomeReferenceCanvas methodsFor: 'private' stamp: 'ar 4/16/2005 23:19'!
transformChanged
	"The current transform has changed. Do whatever is necessary."! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/17/2005 23:11'!
transformDistance: aNumber
	"Transform the given distance from user to device space"
	"@@@@: FIXME - this is incorrect."
	^(ctm transformDirection: aNumber@0) r! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/16/2005 23:09'!
transformPoint: aPoint
	"Transform the given point from user to device space"
	^self transform transformPoint: aPoint! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/17/2005 14:03'!
transform: aTransform
	"Set the transform to be used for subsequent drawing"
	ctm := aTransform.
	self transformChanged.! !

!RomeReferenceCanvas methodsFor: 'transforming' stamp: 'ar 4/16/2005 23:08'!
translateBy: offset
	"Offset all further drawing operations by offset"
	^self transformBy: (COffsetTransform withOffset: offset)! !

!RomeReferenceCanvas methodsFor: 'drawing text' stamp: 'bf 7/31/2006 15:53'!
ttString: aString at: aPoint from: startIndex to: stopIndex
	"Draw a string with the given set of parameters"

	self preserveStateDuring: [
		"we draw the glyphs in their native coordinate system ..."
		pen hasEffect ifTrue: [
			"... however, this must not affect line width."
			pen := pen transformedBy: (CMatrixTransform withScale: font glyphScale / self fontSize)].

		self translateBy: aPoint.	
		self transformBy: fontMatrix.
		self scaleBy: 1.0@-1.0 / font glyphScale. 

		startIndex to: stopIndex do: [:index |
			| advance |
			self drawPath: [advance := font outlineChar: (aString at: index) on: self].
			index < stopIndex ifTrue: [ | kerning |
				kerning := font kerningLeft: (aString at: index) right: (aString at: index + 1).
				self translateBy: advance + (kerning@0.0) ]]].
! !

!RomeTTFont methodsFor: '*Rome-Reference' stamp: 'bf 6/12/2006 18:03'!
outlineChar: aChar on: aCanvas
	| glyph |
	glyph := description at: aChar asUnicode.
	aCanvas bezierShape: glyph contours.
	^glyph advanceWidth @ 0! !

!RomeTTFont methodsFor: '*Rome-Reference' stamp: 'bf 7/31/2006 14:55'!
outlineChar: aChar size: aSize transform: aTransform on: aCanvas
	
	| scale tfm p advance |
	scale := aSize / self glyphScale.
	tfm := aTransform composedWithLocal: (CMatrixTransform withScale: scale @ scale negated).
	p := RomeReferenceCanvas new in: [:tmpCanvas | 
		tmpCanvas makePath: [advance := self outlineChar: aChar on: tmpCanvas]].
	p transformBy: tfm.
	aCanvas path: p.
	^advance / scale

! !

Object subclass: #RomeClipshape
	instanceVariableNames: 'bounds points'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rome-Reference'!

!RomeClipshape methodsFor: 'accessing' stamp: 'ar 4/17/2005 21:45'!
asClipShape
	^self! !

!RomeClipshape methodsFor: 'converting' stamp: 'ar 4/17/2005 21:47'!
asPolygon
	^self points! !

!RomeClipshape methodsFor: 'converting' stamp: 'ar 4/17/2005 21:47'!
asRectangle
	^self bounds! !

!RomeClipshape methodsFor: 'accessing' stamp: 'ar 4/17/2005 21:45'!
bounds
	^bounds ifNil:[self computeBounds]! !

!RomeClipshape methodsFor: 'accessing' stamp: 'ar 4/17/2005 21:45'!
bounds: newBounds
	bounds := newBounds.
	points := nil.! !

!RomeClipshape methodsFor: 'private' stamp: 'ar 4/17/2005 21:47'!
computeBounds
	points size = 0 
		ifTrue:[bounds := 0@0 extent: 0@0]
		ifFalse:[bounds := Rectangle encompassing: points].
	^bounds! !

!RomeClipshape methodsFor: 'private' stamp: 'bf 7/14/2006 20:34'!
computePoints
	| origin corner |
	origin := bounds origin.
	corner := bounds corner.
	^Array
		with: origin
		with: corner x @ origin y
		with: corner
		with: origin x @ corner y
		with: origin
! !

!RomeClipshape methodsFor: 'clipping' stamp: 'ar 4/17/2005 21:46'!
containsPoints: pointList
	pointList do:[:pt| (self containsPoint: pt) ifFalse:[^false]].
	^true! !

!RomeClipshape methodsFor: 'clipping' stamp: 'ar 4/17/2005 21:46'!
containsPoint: aPoint
	| nextPt lastPt dx dy px py |
	(self bounds containsPoint: aPoint) ifFalse:[^false].
	self isRectangular ifTrue:[^true].
	nextPt := points at: 1.
	2 to: points size do:[:i|
		lastPt := nextPt.
		nextPt := points at: i.
		dx := nextPt x - lastPt x.
		dy := nextPt y - lastPt y.
		px := aPoint x - lastPt x.
		py := aPoint y - lastPt y.
		(dx * py) - (dy * px) < 0 ifTrue:[^false].
	].
	^true! !

!RomeClipshape methodsFor: 'clipping' stamp: 'ar 4/17/2005 21:46'!
containsRect: aRectangle
	(self bounds containsRect: aRectangle) ifFalse:[^false].
	self isRectangular ifTrue:[^true].
	(self containsPoint: aRectangle origin) ifFalse:[^false].
	(self containsPoint: aRectangle corner) ifFalse:[^false].
	(self containsPoint: aRectangle topRight) ifFalse:[^false].
	(self containsPoint: aRectangle bottomLeft) ifFalse:[^false].
	^true! !

!RomeClipshape methodsFor: 'clipping' stamp: 'bf 7/14/2006 16:04'!
intersectGeneralPoly: contourList
	| wasCcw ccwList clipped |
	contourList isEmpty ifTrue: [
		^contourList].
	contourList size = 1 ifTrue: [
		^{RomePolygonClipper clip: contourList first against: self points}].
	^(contourList collect: [:pointList |
		wasCcw := (RomePolygonClipper signedAreaOf: pointList) positive.
		ccwList := wasCcw ifTrue: [pointList] ifFalse: [pointList reversed].
		clipped := RomePolygonClipper clip: ccwList against: self points.
		wasCcw ifTrue: [clipped] ifFalse: [clipped reversed]]
	) reject: [:pts | pts isEmpty]! !

!RomeClipshape methodsFor: 'clipping' stamp: 'ar 4/17/2005 21:46'!
intersectPoints: pointList
	^RomeClipshape points: (RomePolygonClipper clip: self points against: pointList).! !

!RomeClipshape methodsFor: 'clipping' stamp: 'bf 7/14/2006 14:58'!
intersectPoly: pointList
	^(RomePolygonClipper clip: pointList against: self points).! !

!RomeClipshape methodsFor: 'clipping' stamp: 'ar 4/17/2005 21:52'!
intersect: aClipShape
	"Compute the intersection of the receiver and aClipShape"
	(self isRectangular and:[aClipShape isRectangular]) ifTrue:[
		^RomeClipshape new bounds: (self bounds intersect: aClipShape bounds).
	] ifFalse:[
		^RomeClipshape new points: 
			(RomePolygonClipper clip: self points against: aClipShape points).
	].! !

!RomeClipshape methodsFor: 'testing' stamp: 'ar 4/17/2005 21:47'!
isRectangular
	"Answer whether this clip shape is rectangular"
	^points == nil! !

!RomeClipshape methodsFor: 'accessing' stamp: 'ar 4/17/2005 21:45'!
points
	^points ifNil:[self computePoints]! !

!RomeClipshape methodsFor: 'accessing' stamp: 'bf 7/17/2006 14:48'!
points: newPoints
	newPoints size = 5 ifTrue: [
		"optimize axis-aligned rectangles"
		| min13 min24 max13 max24 |
		min13 := (newPoints at: 1) min: (newPoints at: 3).
		min24 := (newPoints at: 2) min: (newPoints at: 4).
		min13 = min24 ifTrue: [
			max13 := (newPoints at: 1) max: (newPoints at: 3).
			max24 := (newPoints at: 2) max: (newPoints at: 4).
			max13 = max24
				ifTrue: [^self bounds: (min13 corner: max13)]]].

	points := newPoints.
	bounds := nil.! !

Object subclass: #RomeFill
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rome-Reference'!

!RomeFill methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:14'!
fillOval: aRectangle on: aCanvas
	"By default, clip and fill rectangle"
	aCanvas clipPath:[aCanvas oval: aRectangle].
	self fillRectangle: aRectangle on: aCanvas! !

!RomeFill methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:14'!
fillPath: aPath on: aCanvas
	"By default, clip and fill rectangle"
	aCanvas clipPath: aPath.
	self fillRectangle: aPath bounds on: aCanvas! !

!RomeFill methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:20'!
fillPolygon: aCollection on: aCanvas
	"By default, clip and fill rectangle"
	| path |
	path := aCanvas makePath:[aCanvas polygon: aCollection].
	self fillPath: path on: aCanvas! !

!RomeFill methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:15'!
fillRectangle: aRectangle on: aCanvas
	"You ought to know this at least..."
	^self subclassResponsibility! !

!RomeFill methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:15'!
fillRoundRect: aRectangle radius: radius corners: flags on: aCanvas
	"By default, clip and fill rectangle"
	aCanvas clipPath:[aCanvas roundRect: aRectangle radius: radius corners: flags].
	self fillRectangle: aRectangle on: aCanvas! !

Object subclass: #RomePath
	instanceVariableNames: 'pathData firstPt bounds closed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rome-Reference'!

!RomePath commentStamp: '<historical>' prior: 0!
Instance variables:
	pathData	<OrderedCollection>	The data for the path.
	bounds	<Rectangle>			The bounding box for the path.
	firstPt	<Point>				The start of the last open part, or nil.
	closed	<Boolean>			Wether some part of this path is closed.

The pathData consist of op codes followed by Point parameters:
	0 <to>						move
	1 <to>						line
	2 <via> <to>					quadratic bezier segment
	3 <via> <via> <to>				cubic bezier segment
	4 <first>						close (arg must be equal to first point)
!

!RomePath methodsFor: 'accessing' stamp: 'ar 4/17/2005 19:36'!
bounds
	^bounds ifNil:[bounds := self computeBounds].! !

!RomePath methodsFor: 'accessing' stamp: 'ar 4/17/2005 14:32'!
closed
	"Answer whether this path is closed"
	^closed! !

!RomePath methodsFor: 'creation' stamp: 'jl 7/12/2006 14:58'!
closePath
	"Close this path"
	firstPt ifNotNil: [
		pathData add: 4.
		pathData add: firstPt.
		closed := true].
	firstPt := nil.
! !

!RomePath methodsFor: 'accessing' stamp: 'bf 5/9/2006 16:22'!
computeBounds
	| in op start end via1 via2 min max |
	min := max := nil.
	in := ReadStream on: pathData.
	[in atEnd] whileFalse:[
		op := in next.
		(op > 0 and:[op < 4]) ifFalse:[
			"moveTo, closePath" 
			start := in next.
		] ifTrue:[
			"lineTo, curveTo, curveToVia"
			min ifNil:[min := max := start].
			op = 1 ifTrue:[
				end := in next.
				min := min min: end. max := max max: end.
			].
			op = 2 ifTrue:[
				via1 := in next.
				end := in next.
				min := min min: end. max := max max: end.
				min := min min: via1. max := max max: via1.
			].
			op = 3 ifTrue:[
				via1 := in next.
				via2 := in next.
				end := in next.
				min := min min: end. max := max max: end.
				min := min min: via1. max := max max: via1.
				min := min min: via2. max := max max: via2.
			].
			start := end.
		].
	].
	^min corner: max! !

!RomePath methodsFor: 'accessing' stamp: 'bf 5/18/2006 11:51'!
currentPoint
	^firstPt
		ifNil: [nil]
		ifNotNil: [pathData last]! !

!RomePath methodsFor: 'creation' stamp: 'bf 5/18/2006 13:54'!
curveVia: cp1 and: cp2 to: aPoint
	firstPt ifNil: [self moveTo: 0@0].
	pathData last = aPoint ifFalse: [
		pathData add: 3.
		pathData add: cp1.
		pathData add: cp2.
		pathData add: aPoint.
	].
! !

!RomePath methodsFor: 'creation' stamp: 'bf 5/18/2006 13:54'!
curveVia: cp1 to: aPoint
	firstPt ifNil: [self moveTo: 0@0].
	pathData last = aPoint ifFalse: [
		pathData add: 2.
		pathData add: cp1.
		pathData add: aPoint.
	].
! !

!RomePath methodsFor: 'accessing' stamp: 'bf 6/6/2006 11:54'!
degree
	"answer 0, 1, 2, or 3 depending on the highest-order segment in this path"

	| degree in op |
	degree := 0.
	in := ReadStream on: pathData.
	[in atEnd] whileFalse:[
		op := in next.
		(op > 0 and: [op < 4]) ifTrue:[
			"lineto, curve2, curve3"
			op > degree ifTrue: [
				degree := op.
				degree = 3 ifTrue: [^degree]].
			in skip: op.
		] ifFalse: [
			"moveto, closepath"
			in skip: 1.
		]
	].
	^degree! !

!RomePath methodsFor: 'flattening' stamp: 'bf 5/9/2006 17:32'!
flattenBezier2From: a via: b to: c tolerance: tolerance do: aBlock
	| dist ab bc mid |
	dist := self squaredDistanceOf: b toLineFrom: a to: c.
	dist <= tolerance ifTrue:[^aBlock value: c].
	"Recurse"
	ab := (a + b) * 0.5.
	bc := (b + c) * 0.5.
	mid := (ab + bc) * 0.5. 
	self flattenBezier2From: a via: ab to: mid tolerance: tolerance do: aBlock.
	self flattenBezier2From: mid via: bc to: c tolerance: tolerance do: aBlock.! !

!RomePath methodsFor: 'flattening' stamp: 'bf 5/9/2006 17:40'!
flattenBezier3From: a via: b and: c to: d tolerance: tolerance do: aBlock
	| dist1 dist2 ab bc cd abbc bccd  mid |
	"Check smoothness"
	dist1 := self squaredDistanceOf: b toLineFrom: a to: d.
	dist2 := self squaredDistanceOf: c toLineFrom: a to: d.
	(dist1 max: dist2) <= tolerance ifTrue:[^aBlock value: d].
	"Recurse"
	ab := (a + b) * 0.5.
	bc := (b + c) * 0.5.
	cd := (c + d) * 0.5.
	abbc := (ab + bc) * 0.5.
	bccd := (bc + cd) * 0.5.
	mid := (abbc + bccd) * 0.5.
	self flattenBezier3From: a via: ab and: abbc to: mid tolerance: tolerance do: aBlock.
	self flattenBezier3From: mid via: bccd and: cd to: d tolerance: tolerance do: aBlock.! !

!RomePath methodsFor: 'flattening' stamp: 'jl 7/14/2006 12:31'!
flatten: tolerance
	"Create a polyline for this path"
	| start in op end via1 via2 out poly |
	start := 0@0.
	in := ReadStream on: pathData.
	out := WriteStream on: Array new.
	poly := nil.
	[in atEnd] whileFalse:[
		op := in next.
		(op > 0 and:[op < 4]) ifTrue:[
			poly ifNil:[
				poly := WriteStream on: (Array new: 4). 
				poly nextPut: start.
			].
			op = 1 ifTrue:["lineTo"
				end := in next.
				poly nextPut: end.
			].
			op = 2 ifTrue:["bezier2"
				via1 := in next.
				end := in next.
				self flattenBezier2From: start via: via1 to: end tolerance: tolerance do:[:pt|
					poly nextPut: pt.
				].
			].
			op = 3 ifTrue:["bezier3"
				via1 := in next.
				via2 := in next.
				end := in next.
				self flattenBezier3From: start via: via1 and: via2 to: end tolerance: tolerance do:[:pt|
					poly nextPut: pt.
				].
			].
			start := end.
		] ifFalse:[
			start := in next.
			poly ifNotNil:[
				op = 4 ifTrue:[poly nextPut: start].
				out nextPut: poly contents.
				poly := nil.
			].
		].
	].
	poly ifNotNil:[out nextPut: poly contents].
	^self reorient: out contents! !

!RomePath methodsFor: 'initialize' stamp: 'bf 7/14/2006 20:22'!
initialize
	firstPt := nil.
	pathData := OrderedCollection new.
	closed := false.! !

!RomePath methodsFor: 'testing' stamp: 'ar 4/17/2005 21:38'!
isEmpty
	^pathData isEmpty! !

!RomePath methodsFor: 'creation' stamp: 'bf 6/8/2006 16:34'!
lineBy: delta
	firstPt ifNil: [self moveTo: 0@0].
	^self lineTo: self currentPoint + delta
! !

!RomePath methodsFor: 'creation' stamp: 'bf 5/18/2006 13:53'!
lineTo: aPoint
	firstPt ifNil: [self moveTo: 0@0].
	pathData last = aPoint ifFalse: [
		pathData add: 1.
		pathData add: aPoint.
	]! !

!RomePath methodsFor: 'creation' stamp: 'ar 4/17/2005 21:38'!
moveTo: aPoint
	"Move to aPoint"
	pathData add: 0.
	pathData add: aPoint.
	firstPt := aPoint.! !

!RomePath methodsFor: 'accessing' stamp: 'jl 7/4/2006 16:54'!
pathOn: aCanvas
	"Put me on a canvas"
	| op in args |
	in := ReadStream on: pathData.
	[in atEnd] whileFalse: [
		op := in next.
		args := in next: (#(1 1 2 3 1) at: op+1). "arguments"
		(op < 4) ifTrue:[
			aCanvas
				perform: (#(moveTo: lineTo: curveVia:to: curveVia:and:to:) at: op+1)
				withArguments: args
		] ifFalse: [
			aCanvas closePath.
		].
	].
! !

!RomePath methodsFor: 'flattening' stamp: 'bf 7/14/2006 16:13'!
reorient: polyContours
	| outer |
	outer := RomePolygonClipper outerContourOf: polyContours.
	^(RomePolygonClipper signedAreaOf: outer) positive
		ifTrue: [polyContours]
		ifFalse: [polyContours collect: [:each | each reversed]]! !

!RomePath methodsFor: 'flattening' stamp: 'ar 4/17/2005 20:00'!
squaredDistanceOf: p toLineFrom: pt1 to: pt2
	"Answer the squared distance of p from the line connecting pt1 and pt2"
	| dx dy  pdx pdy u px py |
	dx := pt2 x - pt1 x.
	dy := pt2 y - pt1 y.
	pdx := p x - pt1 x.
	pdy := p y - pt1 y.
	(dx = 0 and:[dy = 0]) ifTrue:[
		^(pdx * pdx) + (pdy * pdy).
	].
	u := ((pdx * dx) + (pdy * dy)) asFloat / ((dx * dx) + (dy * dy)) asFloat.
	u <= 0.0 ifTrue:[^(pdx * pdx) + (pdy * pdy)].
	u >= 1.0 ifTrue:[
		pdx := p x - pt2 x.
		pdy := p y - pt2 y.
		^(pdx * pdx) + (pdy * pdy)
	].
	px := pt1 x + (u * dx).
	py := pt1 y + (u * dy).
	pdx := p x - px.
	pdy := p y - py.
	^(pdx * pdx) + (pdy * pdy)! !

!RomePath methodsFor: 'accessing' stamp: 'ar 4/17/2005 21:39'!
transformBy: tfm
	"Transform me by aTransform"
	| op in out |
	in := ReadStream on: pathData.
	out := WriteStream on: (Array new: pathData size).
	[in atEnd] whileFalse:[
		op := in next.
		out nextPut: op.
		"lineTo, curveTo, curveTo"
		(op > 0 and:[op < 4]) ifTrue:[
			1 to: op do:[:i| out nextPut: (tfm transformPoint: in next)].
		] ifFalse:[
			"moveTo, closePath" 
			out nextPut: (tfm transformPoint: in next)
		].
	].
	pathData := out contents.
	bounds := nil.! !

!RomePath methodsFor: 'accessing' stamp: 'ar 4/17/2005 19:11'!
transformedBy: aTransform
	^self copy transformBy: aTransform! !

Object subclass: #RomePen
	instanceVariableNames: 'width color'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rome-Reference'!

!RomePen class methodsFor: 'instance creation' stamp: 'ar 4/17/2005 19:29'!
width: w color: aColor
	^self new width: w color: aColor! !

!RomePen methodsFor: 'accessing' stamp: 'ar 4/17/2005 20:06'!
color
	^color! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:32'!
drawLineFrom: startPoint to: endPoint on: aCanvas
	^self subclassResponsibility! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:31'!
drawPolyline: aCollection on: aCanvas
	"Bail to lines"
	| prevPt |
	prevPt := aCollection first.
	aCollection do:[:nextPt|
		nextPt = prevPt ifFalse:[self drawLineFrom: prevPt to: nextPt on: aCanvas].
		prevPt := nextPt.
	].! !

!RomePen methodsFor: 'accessing' stamp: 'bf 7/31/2006 17:01'!
expandBounds: aRectangle
	^aRectangle expandByRomePenWidth: width color: color! !

!RomePen methodsFor: 'accessing' stamp: 'ar 4/17/2005 22:51'!
fill
	^color! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:28'!
frameOval: aRectangle on: aCanvas
	^self framePath: (aCanvas makePath:[aCanvas oval: aRectangle]) on: aCanvas! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:29'!
framePath: aPath on: aCanvas
	"Bail to polygon"
	(aPath flatten: aCanvas deviceTolerance) do:[:poly|
		self framePolygon: poly on: aCanvas
	].! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:30'!
framePolygon: aCollection on: aCanvas
	"Bail to lines"
	| prevPt |
	prevPt := aCollection last.
	aCollection do:[:nextPt|
		nextPt = prevPt ifFalse:[self drawLineFrom: prevPt to: nextPt on: aCanvas].
		prevPt := nextPt.
	].! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:28'!
frameRectangle: aRectangle on: aCanvas
	"Bail to path"
	^self framePath: (aCanvas makePath:[aCanvas rectangle: aRectangle]) on: aCanvas! !

!RomePen methodsFor: 'drawing' stamp: 'ar 4/17/2005 21:28'!
frameRoundRect: aRectangle radius: radius corners: flags on: aCanvas
	"Bail to path"
	^self framePath: (aCanvas makePath:[
		aCanvas roundRect: aRectangle radius: radius corners: flags
	]) on: aCanvas! !

!RomePen methodsFor: 'testing' stamp: 'bf 6/8/2006 14:43'!
hasEffect
	^width > 0 and: [color isTransparent not]! !

!RomePen methodsFor: 'accessing' stamp: 'bf 7/31/2006 17:01'!
insetBounds: aRectangle
	^aRectangle insetByRomePenWidth: width color: color! !

!RomePen methodsFor: 'testing' stamp: 'ar 4/17/2005 21:39'!
isSolid
	^color isSolidFill! !

!RomePen methodsFor: 'printing' stamp: 'bf 6/23/2006 15:21'!
printOn: aStream
	aStream
		nextPutAll: self class name;
		nextPut: $(;
		print: width;
		space;
		print: color;
		nextPut: $)! !

!RomePen methodsFor: 'transforming' stamp: 'ar 4/17/2005 19:42'!
transformBy: aTransform
	width := (aTransform transformDirection: 0@width) r truncated.
	color isColor ifFalse:[
		color := color transformedBy: aTransform.
	].! !

!RomePen methodsFor: 'transforming' stamp: 'ar 4/17/2005 19:38'!
transformedBy: aTransform
	"Transform me into device space"
	^self copy transformBy: aTransform! !

!RomePen methodsFor: 'accessing' stamp: 'ar 4/17/2005 20:06'!
width
	^width! !

!RomePen methodsFor: 'initialize' stamp: 'ar 4/17/2005 19:29'!
width: w color: aColor
	width := w.
	color := aColor.! !

Object subclass: #RomePolygonClipper
	instanceVariableNames: ''
	classVariableNames: 'Debug'
	poolDictionaries: ''
	category: 'Rome-Reference'!

!RomePolygonClipper commentStamp: '<historical>' prior: 0!
CPolygonClipper new halt
	intersect: 0@50 dir: 100@0
	with: 200@0 dir: 0@200.!

!RomePolygonClipper class methodsFor: 'instance creation' stamp: 'ar 4/18/2004 17:25'!
clip: polyPath against: clipPath
	^self new clip: polyPath against: clipPath! !

!RomePolygonClipper class methodsFor: 'accessing' stamp: 'ar 4/18/2004 15:36'!
debug
	^Debug! !

!RomePolygonClipper class methodsFor: 'accessing' stamp: 'ar 4/18/2004 15:36'!
debug: aBool
	Debug := aBool.! !

!RomePolygonClipper class methodsFor: 'class initialization' stamp: 'ar 4/19/2004 02:25'!
initialize
	"CPolygonClipper initialize"
	Debug := false.! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'bf 4/7/2006 15:44'!
makeStarPath: npts
	| pt ext oldR points pts |
	pt := 100 @ 100.
	ext := pt r.
	oldR := ext.
	points := npts.
	pts := (0 to: 360 by: 360 // points // 2)
				collect: [:angle | (Point r: (oldR := oldR = ext
									ifTrue: [ext * 5 // 12]
									ifFalse: [ext]) degrees: angle + pt degrees)
						+ (150 @ 150)].
	"Close path properly"
	pts at: 1 put: pts last.
	^pts! !

!RomePolygonClipper class methodsFor: 'utilities' stamp: 'bf 7/14/2006 16:12'!
outerContourOf: contourList
	| top |
	contourList size = 1 ifTrue: [^contourList first].
	^contourList detectMax: [:points | 
		top := points first y.
		2 to: points size do: [:i | 
			(points at: i) y > top ifTrue: [top := (points at: i) y]].
		top].! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:42'!
profileClipStar
	"RomePolygonClipper profileClipStar"
	| path1 path2 path3 clip |
	Debug := false.
	path1 := {50@50. 250@50. 250@250. 50@250. 50@50}.
	path2 := {150@0. 300@150. 150@300. 0@150. 150@0}.
	path3 := (RomePolygonClipper makeStarPath: 180) collect:[:each| each asIntegerPoint].

	clip := RomePolygonClipper clip: path1 against: path2.
	MessageTally spyOn:[
		1 to: 1000 do:[:i| RomePolygonClipper clip: path3 against: clip]]! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:43'!
profileClipStarInside
	"RomePolygonClipper profileClipStarInside"
	| path3 clip |
	Debug := false.
	clip := {0@0. 300@0. 300@300. 0@300. 0@0}.
	path3 := (RomePolygonClipper makeStarPath: 180) collect:[:each| each asIntegerPoint].
	MessageTally spyOn:[
		1 to: 1000 do:[:i| RomePolygonClipper clip: path3 against: clip]]
! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:42'!
profileClipStarOutside
	"RomePolygonClipper profileClipStarOutside"
	| path3 clip |
	Debug := false.
	clip := {0@0. 300@0. 300@300. 0@300. 0@0} collect:[:pt| pt + 300].
	path3 := (RomePolygonClipper makeStarPath: 180) collect:[:each| each asIntegerPoint].
	MessageTally spyOn:[
		1 to: 1000 do:[:i| RomePolygonClipper clip: path3 against: clip]]
! !

!RomePolygonClipper class methodsFor: 'utilities' stamp: 'bf 7/14/2006 15:38'!
signedAreaOf: points
	| area a b |
	area := 0.
	1 to: points size do: [:i |
		a := points at: i.
		b := points atWrap: i+1.
		area := area + (a x * b y) - (b x * a y).
	].
	^area / 2.0! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:43'!
testClipStar
	"RomePolygonClipper testClipStar"
	| path1 path2 path3 clip |
	Debug := true.
	path1 := {50@50. 250@50. 250@250. 50@250. 50@50}.
	path2 := {150@0. 300@150. 150@300. 0@150. 150@0}.
	path3 := (RomePolygonClipper makeStarPath: 180) collect:[:each| each asIntegerPoint].

	clip := RomePolygonClipper clip: path1 against: path2.
	RomePolygonClipper clip: path3 against: clip! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:42'!
testClipStarInside
	"RomePolygonClipper testClipStarInside"
	| path3 clip |
	Debug := true.
	clip := {0@0. 300@0. 300@300. 0@300. 0@0}.
	path3 := (RomePolygonClipper makeStarPath: 180) collect:[:each| each asIntegerPoint].
	RomePolygonClipper clip: path3 against: clip! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:42'!
testClipStarOutside
	"RomePolygonClipper testClipStarOutside"
	| path3 clip |
	Debug := true.
	clip := {0@0. 300@0. 300@300. 0@300. 0@0} collect:[:pt| pt + 300].
	path3 := (RomePolygonClipper makeStarPath: 180) collect:[:each| each asIntegerPoint].
	RomePolygonClipper clip: path3 against: clip! !

!RomePolygonClipper class methodsFor: 'examples' stamp: 'ar 4/17/2005 21:42'!
testClipStarPartwayOutside
	"RomePolygonClipper testClipStarPartwayOutside"
	| path3 clip |
	Debug := true.
	clip := {0@0. 300@0. 300@300. 0@300. 0@0} collect:[:pt| pt + 200].
	path3 := (RomePolygonClipper makeStarPath: 36) collect:[:each| each asIntegerPoint].
	^RomePolygonClipper clip: path3 against: clip! !

!RomePolygonClipper methodsFor: 'clipping' stamp: 'ar 4/20/2004 21:00'!
clip: polyPath against: clipPath
	| lastClip nextClip path |
	clipPath size < 3 ifTrue:[^#()].
	polyPath size < 3 ifTrue:[^#()].
	clipPath first = clipPath last ifFalse:[self error: 'Clippath not closed'].
	Debug ifTrue:[self debugDrawPath: Color black points: clipPath].
	Debug ifTrue:[self debugDrawPath: Color blue points: polyPath].
	path := polyPath.
	path first = path last ifFalse:[path := path copyWith: path first].
	nextClip := clipPath at: 1.
	2 to: clipPath size do:[:clipIndex|
		lastClip := nextClip.
		nextClip := clipPath at: clipIndex.
		path := self clip: path from: lastClip to: nextClip.
		path size < 3 ifTrue:[^#()].
	].
	Debug ifTrue:[self debugDrawPath: Color green points: path].
	^path! !

!RomePolygonClipper methodsFor: 'clipping' stamp: 'ar 4/18/2004 17:24'!
clip: path from: lastClip to: nextClip
	| lastClipX lastClipY clipX clipY index nextPt dx dy nextSide inside lastPt lastSide result pt lastInside |

	"Allocate result lazily - this is good for paths which are fully inside or outside of the clip edge since we won't ever allocate temporary storage for them."
	result := nil.

	"Cache lastClip's values"
	lastClipX := lastClip x.
	lastClipY := lastClip y.

	"Compute slope of the next clip segment"
	clipX := nextClip x - lastClipX.
	clipY := nextClip y - lastClipY.

	"Find first point which isn't on contour to determine inside/outside"
	index := 0.
	[nextPt := path at: (index := index + 1).
	dx := nextPt x - lastClipX.
	dy := nextPt y - lastClipY.
	"The cross product hands us either one of <0/0/>0 depending on the side"
	nextSide := (clipX * dy) - (dx * clipY).
	nextSide = 0] whileTrue:[
		"Check for degenerate case of all points being on the boundary of the clip edge"
		index = path size ifTrue:[^#()].
	].
	inside := nextSide > 0.
	inside ifTrue:[lastInside := index].

	"Now go over the path and see where we cross the boundaries"
	[index < path size] whileTrue:[
		lastPt := nextPt.
		lastSide := nextSide.

		"Compute location of nextPt relative to clip edge"
		nextPt := path at: (index := index + 1).
		dx := nextPt x - lastClipX.
		dy := nextPt y - lastClipY.

		"The cross product hands us either one of <0/0/>0 depending on the side"
		nextSide := (clipX * dy) - (dx * clipY).

		"If we run on the clip edge keep the last side for reference"
		nextSide = 0 ifTrue:[nextSide := lastSide].

		(nextSide sign * lastSide sign < 0) ifTrue:[
			"We're crossing the clip edge -- compute the intersection point"
			dx := nextPt x - lastPt x.
			dy := nextPt y - lastPt y.
			pt := self intersect: lastClip dir: clipX@clipY with: lastPt dir: dx@dy.
			result ifNil:[result := WriteStream on: (Array new: 4)].
			inside ifTrue:[
				"We're leaving the clip path - copy result into output"
				result nextPutAll: (path copyFrom: lastInside to: index-1).

				"Note: To avoid duplicates in the out path we only remember
				the intersection point if it doesn't coincide with the start point
				of the leaving edge (which is a visible end-point so we have
				recorded it above)"
				(pt x = lastPt x and:[pt y = lastPt y])
					ifFalse:[result nextPut: pt].
			] ifFalse:[
				"We're entering the clip path"
				result nextPut: pt.
			].
			Debug ifTrue:[self debugDrawPoint: Color red at: pt].
			inside := inside not.
			inside ifTrue:[lastInside := index].
		].
	].
	inside ifTrue:[
		result ifNil:[^path].
		"Copy remainder of visible points"
		result nextPutAll: (path copyFrom: lastInside to: index).
	] ifFalse:[
		result ifNil:[^#()].
		"Close path again if needed"
		result position > 0 ifTrue:[result nextPut: result originalContents first].
	].
	^result contents! !

!RomePolygonClipper methodsFor: 'debug' stamp: 'ar 4/18/2004 04:14'!
debugDrawLine: aColor from: start to: stop
	Display getCanvas line: start*2 to: stop*2 width: 2 color: aColor.
	(Delay forMilliseconds: 50) wait.! !

!RomePolygonClipper methodsFor: 'debug' stamp: 'jl 7/12/2006 16:38'!
debugDrawPath: aColor points: ptList
	| lastPt nextPt |
	lastPt := ptList first // 64.
	Display getCanvas fillRectangle: (lastPt*2-2 extent: 4@4) color: aColor.
	2 to: ptList size do:[:i|
		nextPt := (ptList at: i) // 64.
		Display getCanvas fillRectangle: (nextPt*2-2 extent: 4@4) color: aColor.
		Display getCanvas line: lastPt*2 to: nextPt*2 width: 2 color: aColor.
		lastPt := nextPt.
	].
	(Delay forMilliseconds: 200) wait.
! !

!RomePolygonClipper methodsFor: 'debug' stamp: 'ar 4/18/2004 04:23'!
debugDrawPoint: aColor at: pt
	Display getCanvas fillRectangle: (pt*2-2 extent: 5@5) color: aColor.
	(Delay forMilliseconds: 50) wait.! !

!RomePolygonClipper methodsFor: 'clipping' stamp: 'bf 4/7/2006 15:44'!
intersect: pt1 dir: dir1 with: pt2 dir: dir2
	"Compute the intersection of the line starting at pt1 with the line starting at pt2"
	| det alpha deltaX deltaY |
	det := (dir1 x * dir2 y) - (dir1 y * dir2 x).
	deltaX := pt2 x - pt1 x.
	deltaY := pt2 y - pt1 y.
	alpha := (deltaX * dir2 y) - (deltaY * dir2 x).
	"And compute intersection"
	^(pt1 x + (alpha * dir1 x // det)) @ (pt1 y + (alpha * dir1 y // det))
! !
RomePolygonClipper initialize!
